from .index_collections import IndexCollection, FeatureIndexCollection, MultiLabelIndexCollection, map_whole_index_to_train
from .multi_label_tools import check_index_multilabel, infer_label_size_multilabel, flattern_multilabel_index, integrate_multilabel_index, get_labelmatrix_in_multilabel, get_Xy_in_multilabel